local function FixScrollOnUpdate(frame)
	frame:SetScript("OnUpdate", nil)
	frame.obj:FixScroll()
end

--[[-----------------------------------------------------------------------------
Scripts
-------------------------------------------------------------------------------]]

local function Button_OnClick(frame)
	PlaySound("gsTitleOptionExit")
	frame:GetParent():Hide()
end

local function Title_OnMouseDown(frame)
	frame:GetParent():StartMoving()
end

local function MoverSizer_OnMouseUp(mover)
	mover:GetParent():StopMovingOrSizing()
end

local function SizerSE_OnMouseDown(frame)
	frame:GetParent():StartSizing("BOTTOMRIGHT")
end

local function SizerS_OnMouseDown(frame)
	frame:GetParent():StartSizing("BOTTOM")
end

local function SizerE_OnMouseDown(frame)
	frame:GetParent():StartSizing("RIGHT")
end

local function ScrollFrame_OnMouseWheel(frame, value)
	frame.obj:MoveScroll(value)
end

local function ScrollFrame_OnSizeChanged(frame)
	frame:SetScript("OnUpdate", FixScrollOnUpdate)
end

local function ScrollBar_OnScrollValueChanged(frame, value)
	frame.obj:SetScroll(value)
end

local function CheckBox_Update(frame)
	if ARPC_Collapse then
		SetDesaturation(frame.obj.check, false)
		frame.obj.check:Show()
		PlaySound("igMainMenuOptionCheckBoxOn")
	else
		SetDesaturation(frame.obj.check, false)
		frame.obj.check:Hide()
		PlaySound("igMainMenuOptionCheckBoxOff")
	end
end

local function CheckBox_OnMouseDown(frame)
	frame.obj.checktext:SetPoint("LEFT", frame.obj.checkbg, "RIGHT", 1, -1)
end

local function CheckBox_OnMouseUp(frame)
	ARPC_Collapse = not ARPC_Collapse

	frame.obj.checktext:SetPoint("LEFT", frame.obj.checkbg, "RIGHT")
	frame.obj.checktext:SetPoint("RIGHT")

	CheckBox_Update(frame)

	ARPC:UpdateAllVisibilities()
end

local function SetButton_OnClick(frame)
	ARPC:ShowPointsPopup()
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {
	["Toggle"] = function(self)
		if self.frame:IsShown() then 
			self.frame:Hide() 
		else 
			self.frame:Show() 
		end
	end,

	["SetScale"] = function(self, value)
		ARPC_Scale = value
		self.frame:SetScale(value)
	end,
	
	["SetScroll"] = function(self, value)
		local status = self.status or self.localstatus
		local viewheight = self.scrollframe:GetHeight()
		local height = self.content:GetHeight()
		local offset

		if viewheight > height then
			offset = 0
		else
			offset = floor((height - viewheight) / 1000.0 * value)
		end
		self.content:ClearAllPoints()
		self.content:SetPoint("TOPLEFT", 0, offset)
		self.content:SetPoint("TOPRIGHT", 0, offset)
		status.offset = offset
		status.scrollvalue = value
	end,

	["MoveScroll"] = function(self, value)
		local status = self.status or self.localstatus
		local height, viewheight = self.scrollframe:GetHeight(), self.content:GetHeight()
		
		if self.scrollBarShown then
			local diff = height - viewheight
			local delta = 1
			if value < 0 then
				delta = -1
			end
			self.scrollbar:SetValue(min(max(status.scrollvalue + delta*(1000/(diff/45)),0), 1000))
		end
	end,

	["FixScroll"] = function(self)
		if self.updateLock then return end
		self.updateLock = true
		local status = self.status or self.localstatus
		local height, viewheight = self.scrollframe:GetHeight(), self.content:GetHeight()
		local offset = status.offset or 0
		local curvalue = self.scrollbar:GetValue()
		-- Give us a margin of error of 2 pixels to stop some conditions that i would blame on floating point inaccuracys
		-- No-one is going to miss 2 pixels at the bottom of the frame, anyhow!
		if viewheight < height + 2 then
			if self.scrollBarShown then
				self.scrollBarShown = nil
				self.scrollbar:Hide()
				self.scrollbar:SetValue(0)
				self.scrollframe:SetPoint("BOTTOMRIGHT")
				self:DoLayout()
			end
		else
			if not self.scrollBarShown then
				self.scrollBarShown = true
				self.scrollbar:Show()
				self.scrollframe:SetPoint("BOTTOMRIGHT", -20, 0)
				self:DoLayout()
			end
			local value = (offset / (viewheight - height) * 1000)
			if value > 1000 then value = 1000 end
			self.scrollbar:SetValue(value)
			self:SetScroll(value)
			if value < 1000 then
				self.content:ClearAllPoints()
				self.content:SetPoint("TOPLEFT", 0, offset)
				self.content:SetPoint("TOPRIGHT", 0, offset)
				status.offset = offset
			end
		end
		self.updateLock = nil
	end,
	
	["AddChild"] = function(self, widget)
		tinsert(self.children, widget)
		widget:SetParent(self.content)
		self:DoLayout()
	end,
	
	["RemoveChild"] = function(self, widget)
		for index,entry in ipairs(self.children) do
			if entry == widget then
				table.remove(self.children, index)
				entry:ClearAllPoints()
				entry:Hide()
				entry:SetParent(UIParent)
				break
			end
		end
		self:DoLayout()
	end,
	
	["DoLayout"] = function(self)
		local height = 0
		local width = self.content:GetWidth() or 0
		for i = 1, #self.children do
			local child = self.children[i]
			
			child:ClearAllPoints()
			child:Show()
			if i == 1 then
				child:SetPoint("TOPLEFT", self.content)
			else
				child:SetPoint("TOPLEFT", self.children[i-1], "BOTTOMLEFT")
			end
			
			child:SetWidth(width)
			child:SetPoint("RIGHT", self.content)
			
			height = height + (child:GetHeight() or 0)
		end
		self.content:SetHeight(height or 0 + 20)
		self.scrollframe:SetScript("OnUpdate", FixScrollOnUpdate)
	end
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local FrameBackdrop = {
	bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
	edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
	tile = true, tileSize = 32, edgeSize = 32,
	insets = { left = 8, right = 8, top = 8, bottom = 8 }
}

local PaneBackdrop  = {
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true, tileSize = 16, edgeSize = 16,
	insets = { left = 3, right = 3, top = 5, bottom = 3 }
}

function ARPC_CreateWindow()
	local frame = CreateFrame("Frame", "ARPC_Window", UIParent)
	frame:Hide()

	frame:SetSize(450, 650)
	frame:SetBackdrop(FrameBackdrop)
	frame:SetPoint("CENTER",0,0)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetUserPlaced(true)
	frame:SetResizable(true)
	frame:SetFrameStrata("HIGH")
	frame:SetFrameLevel(1)
	frame:SetBackdropColor(0, 0, 0, 1)
	frame:SetMinResize(400, 200)
	frame:SetToplevel(true)
	if ARPC_Scale ~= nil then frame:SetScale(ARPC_Scale) end
	tinsert(UISpecialFrames, "ARPC_Window")

	local closebutton = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
	closebutton:SetScript("OnClick", Button_OnClick)
	closebutton:SetPoint("BOTTOMRIGHT", -27, 17)
	closebutton:SetHeight(20)
	closebutton:SetWidth(100)
	closebutton:SetText(CLOSE)

	local statusbg = CreateFrame("Button", nil, frame)
	statusbg:SetPoint("BOTTOMLEFT", 15, 15)
	statusbg:SetPoint("BOTTOMRIGHT", -132, 15)
	statusbg:SetHeight(24)
	statusbg:SetBackdrop(PaneBackdrop)
	statusbg:SetBackdropColor(0.1,0.1,0.1)
	statusbg:SetBackdropBorderColor(0.4,0.4,0.4)

	local statustext = statusbg:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	statustext:SetPoint("TOPLEFT", 7, -2)
	statustext:SetPoint("BOTTOMRIGHT", -7, 2)
	statustext:SetHeight(20)
	statustext:SetJustifyH("LEFT")
	statustext:SetText("")

	local titlebg = frame:CreateTexture(nil, "OVERLAY")
	titlebg:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
	titlebg:SetTexCoord(0.31, 0.67, 0, 0.63)
	titlebg:SetPoint("TOP", 0, 12)
	titlebg:SetWidth(100)
	titlebg:SetHeight(40)

	local title = CreateFrame("Frame", nil, frame)
	title:EnableMouse(true)
	title:SetScript("OnMouseDown", Title_OnMouseDown)
	title:SetScript("OnMouseUp", MoverSizer_OnMouseUp)
	title:SetAllPoints(titlebg)

	local titletext = title:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	titletext:SetPoint("TOP", titlebg, "TOP", 0, -14)
	titletext:SetText("Angry Raid Points")
	titlebg:SetWidth((titletext:GetWidth() or 0) + 10)

	local titlebg_l = frame:CreateTexture(nil, "OVERLAY")
	titlebg_l:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
	titlebg_l:SetTexCoord(0.21, 0.31, 0, 0.63)
	titlebg_l:SetPoint("RIGHT", titlebg, "LEFT")
	titlebg_l:SetWidth(30)
	titlebg_l:SetHeight(40)

	local titlebg_r = frame:CreateTexture(nil, "OVERLAY")
	titlebg_r:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
	titlebg_r:SetTexCoord(0.67, 0.77, 0, 0.63)
	titlebg_r:SetPoint("LEFT", titlebg, "RIGHT")
	titlebg_r:SetWidth(30)
	titlebg_r:SetHeight(40)

	local sizer_se = CreateFrame("Frame", nil, frame)
	sizer_se:SetPoint("BOTTOMRIGHT")
	sizer_se:SetWidth(25)
	sizer_se:SetHeight(25)
	sizer_se:EnableMouse()
	sizer_se:SetScript("OnMouseDown",SizerSE_OnMouseDown)
	sizer_se:SetScript("OnMouseUp", MoverSizer_OnMouseUp)

	local line1 = sizer_se:CreateTexture(nil, "BACKGROUND")
	line1:SetWidth(14)
	line1:SetHeight(14)
	line1:SetPoint("BOTTOMRIGHT", -8, 8)
	line1:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
	local x = 0.1 * 14/17
	line1:SetTexCoord(0.05 - x, 0.5, 0.05, 0.5 + x, 0.05, 0.5 - x, 0.5 + x, 0.5)

	local line2 = sizer_se:CreateTexture(nil, "BACKGROUND")
	line2:SetWidth(8)
	line2:SetHeight(8)
	line2:SetPoint("BOTTOMRIGHT", -8, 8)
	line2:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
	local x = 0.1 * 8/17
	line2:SetTexCoord(0.05 - x, 0.5, 0.05, 0.5 + x, 0.05, 0.5 - x, 0.5 + x, 0.5)

	local sizer_s = CreateFrame("Frame", nil, frame)
	sizer_s:SetPoint("BOTTOMRIGHT", -25, 0)
	sizer_s:SetPoint("BOTTOMLEFT")
	sizer_s:SetHeight(25)
	sizer_s:EnableMouse(true)
	sizer_s:SetScript("OnMouseDown", SizerS_OnMouseDown)
	sizer_s:SetScript("OnMouseUp", MoverSizer_OnMouseUp)

	local sizer_e = CreateFrame("Frame", nil, frame)
	sizer_e:SetPoint("BOTTOMRIGHT", 0, 25)
	sizer_e:SetPoint("TOPRIGHT")
	sizer_e:SetWidth(25)
	sizer_e:EnableMouse(true)
	sizer_e:SetScript("OnMouseDown", SizerE_OnMouseDown)
	sizer_e:SetScript("OnMouseUp", MoverSizer_OnMouseUp)
	
	local windowcontent = CreateFrame("Frame", nil, frame)
	windowcontent:SetPoint("TOPLEFT", 17, -27)
	windowcontent:SetPoint("BOTTOMRIGHT", -17, 40)

	local scrollframe = CreateFrame("ScrollFrame", nil, windowcontent)
	scrollframe:SetPoint("TOPLEFT", 0, 0)
	scrollframe:EnableMouseWheel(true)
	scrollframe:SetScript("OnMouseWheel", ScrollFrame_OnMouseWheel)
	scrollframe:SetScript("OnSizeChanged", ScrollFrame_OnSizeChanged)

	local scrollbar = CreateFrame("Slider", "ARPC_ScrollBar", scrollframe, "UIPanelScrollBarTemplate")
	scrollbar:SetPoint("TOPLEFT", scrollframe, "TOPRIGHT", 4, -16)
	scrollbar:SetPoint("BOTTOMLEFT", scrollframe, "BOTTOMRIGHT", 4, 16)
	scrollbar:SetMinMaxValues(0, 1000)
	scrollbar:SetValueStep(1)
	scrollbar:SetValue(0)
	scrollbar:SetWidth(16)
	scrollbar:Hide()
	scrollbar:SetScript("OnValueChanged", ScrollBar_OnScrollValueChanged)

	local scrollbg = scrollbar:CreateTexture(nil, "BACKGROUND")
	scrollbg:SetAllPoints(scrollbar)
	scrollbg:SetTexture(0, 0, 0, 0.4)
	
	local content = CreateFrame("Frame", nil, scrollframe)
	content:SetPoint("TOPLEFT")
	content:SetPoint("TOPRIGHT")
	content:SetHeight(400)
	scrollframe:SetScrollChild(content)

	local checkframe = CreateFrame("Button", nil, statusbg)
	checkframe:SetPoint("BOTTOMRIGHT", -2, 5)
	checkframe:SetWidth(80)
	checkframe:SetHeight(19)
	checkframe:SetScript("OnMouseDown", CheckBox_OnMouseDown)
	checkframe:SetScript("OnMouseUp", CheckBox_OnMouseUp)
	checkframe:EnableMouse(true)

	local checkbg = checkframe:CreateTexture(nil, "ARTWORK")
	checkbg:SetWidth(24)
	checkbg:SetHeight(24)
	checkbg:SetPoint("TOPLEFT")
	checkbg:SetTexture("Interface\\Buttons\\UI-CheckBox-Up")

	local check = checkframe:CreateTexture(nil, "OVERLAY")
	check:SetAllPoints(checkbg)
	check:SetTexture("Interface\\Buttons\\UI-CheckBox-Check")

	local checktext = checkframe:CreateFontString(nil, "OVERLAY", "GameFontHighlight")
	checktext:SetJustifyH("LEFT")
	checktext:SetHeight(18)
	checktext:SetText("Collapse")
	checktext:SetPoint("LEFT", checkbg, "RIGHT")
	checktext:SetPoint("RIGHT")

	local checkhighlight = checkframe:CreateTexture(nil, "HIGHLIGHT")
	checkhighlight:SetTexture("Interface\\Buttons\\UI-CheckBox-Highlight")
	checkhighlight:SetBlendMode("ADD")
	checkhighlight:SetAllPoints(checkbg)

	local widget = {
		localstatus = {},
		children = {},
		
		titletext   = titletext,
		statustext  = statustext,
		statusbg    = statusbg,
		titlebg     = titlebg,
		sizer_se    = sizer_se,
		sizer_s     = sizer_s,
		sizer_e     = sizer_e,
		
		content     = content,
		scrollframe = scrollframe,
		scrollbar   = scrollbar,
		
		checkframe  = checkframe,
		checktext   = checktext,
		checkbg     = checkbg,
		check       = check,

		frame       = frame
	}
	for method, func in pairs(methods) do
		widget[method] = func
	end
	closebutton.obj, statusbg.obj, scrollframe.obj, scrollbar.obj, checkframe.obj = widget, widget, widget, widget, widget

	widget:FixScroll()
	CheckBox_Update(checkframe)

	return widget
end
